#!/usr/bin/perl
#
# Periodically scan the list of available wireless networks and adjust
# the ESSID in the card according to a preferred-SSID list.
#
# Nickolai Zeldovich <kolya@MIT.EDU>
#
# $Id: wifiassocd.pl,v 1.10 2004/04/07 00:18:36 kolya Exp $

use utf8;
use strict;

my $CONFIG = "/etc/wifiassocd.conf";
my $DEVICE;
my $PERIOD;
my @PREFER;
my $DEBUG = 0;

foreach my $arg (@ARGV) {
    if ($arg eq '-d') {
	$DEBUG = 1;
    } else {
	print	"Unknown argument: $arg\n",
		"\n",
		"Usage: wifiassocd.pl [-d]\n",
		"  -d    Enable debug mode\n";
	exit 1;
    }
}

exit if !$DEBUG && fork();

my $lastssid;
my $down = 1;
while (1) {
    read_config();

    my $associated = get_assoc_ap();
    if (!$down && $associated ne 'FF:FF:FF:FF:FF:FF' &&
		  $associated ne '00:00:00:00:00:00' &&
		  $associated ne 'unknown') {
	print "Still associated, sleeping for $PERIOD seconds\n" if $DEBUG;
	sleep $PERIOD;
	next;
    }

    my $newssid;
    my @availssid = scan();
    print "SSID list: ", join(', ', @availssid), "\n" if $DEBUG;

    foreach my $ssid (@PREFER) {
	if (grep {/^$ssid$/} @availssid) {
	    print "Found preferred SSID $ssid\n" if $DEBUG;
	    $newssid = $ssid;
	    last;
	}
	print "Preferred SSID $ssid not found in scan\n" if $DEBUG;
    }

    if (!(defined $newssid) && (defined $availssid[0])) {
	$newssid = $availssid[0];
	print "No preferred SSID found, guessing: $newssid\n" if $DEBUG;
	append_ssid($newssid);
    }

    if (defined $newssid) {
	if ($newssid ne $lastssid) {
	    print "Changing association from old <$lastssid> to ",
		  "new <$newssid>\n" if $DEBUG;
	    $lastssid = $newssid;
	    xsystem("ifdown $DEVICE >/dev/null 2>&1");
	    change_ssid($DEVICE, $newssid);
	    xsystem("ifup $DEVICE >/dev/null 2>&1");
	} else {
	    print "Still associated to $newssid\n" if $DEBUG;
	}
	$down = 0;
    } else {
	print "No available SSIDs, down'ing the interface\n" if $DEBUG;
	xsystem("ifdown $DEVICE");
	$down = 1;
    }

    print "Done with scan, sleeping for $PERIOD seconds\n\n" if $DEBUG;
    sleep $PERIOD;
}

sub get_assoc_ap {
    my $out = `iwconfig $DEVICE`;
    return $1 if $out =~ /Access Point: (..:..:..:..:..:..)/;
    print "Error: unable to determine associated access point for $DEVICE\n"
	if $DEBUG;
    return "unknown";
}

sub scan {
    my @ssid = ();
    xsystem("ifconfig $DEVICE up");
    open(IWSCAN, "iwlist $DEVICE scan |");
    while (<IWSCAN>) {
	push @ssid, $1 if /ESSID:"([^"]*)"$/;
    }
    return @ssid;
}

sub read_config {
    $DEVICE = '';
    $PERIOD = '';
    @PREFER = ();

    open(F, "$CONFIG");
    while (<F>) {
	s/#.*$//;
	s/\s*$//;
	$DEVICE = $1 if /device (.*)/;
	$PERIOD = $1 if /period (.*)/;
	push @PREFER, &url_unquote($1) if /ssid \"(.*)\"/;
    }
    close(F);

    if ($DEBUG) {
	print	"Configuration:\n",
		"  Device: $DEVICE\n",
		"  Period: $PERIOD\n",
		"  SSIDs: ", join(', ', @PREFER), "\n";
    }
}

sub append_ssid {
    my ($ssid) = @_;
    open(F, ">>$CONFIG");
    print F
	"# ", `date`,
	"# Automatically added by wifiassocd\n",
	"ssid \"", &url_quote($ssid), "\"\n";
    close(F);
}

sub change_ssid {
    my ($dev, $ssid) = @_;

    # Rewrite /etc/sysconfig/network-scripts/ifcfg-$dev to have ESSID=$ssid
    my $ifcfg = "/etc/sysconfig/network-scripts/ifcfg-$dev";

    open(I, "$ifcfg");
    open(O, ">$ifcfg.new");

    my $wrotessid = 0;
    while (<I>) {
	if (/^ESSID=/) {
	    print O "ESSID=\"$ssid\"\n";
	    $wrotessid = 1;
	} else {
	    print O;
	}
    }
    if (!$wrotessid) {
	print O "ESSID=\"$ssid\"\n";
    }
    close(O);
    xsystem("mv $ifcfg.new $ifcfg");

    # Redhat sucks
    xsystem("find /etc/sysconfig/networking -name ifcfg-$dev | xargs rm");
}

sub xsystem {
    my (@args) = @_;

    print "Running command: ", join(' ', @args), "\n" if $DEBUG;
    system(@args);
}

sub url_quote {
    my $s = shift;
                                                                                
    utf8::upgrade($s) unless $^V and $^V lt v5.8.0;
                                                                                
    $s =~ s/([^a-zA-Z0-9_.\*~-])/sprintf("%%%02X",ord($1))/eg;
                                                                                
    return $s;
}

sub url_unquote {
    my $out = shift;
                                                                                
    $out =~ s/\+/ /g;
    $out =~ s/%([[:xdigit:]][[:xdigit:]])/ chr hex($1)/ge;
 
    utf8::downgrade($out) unless $^V and $^V lt v5.8.0;
                                                                                
    return $out;
}
